from enum import Enum
from math import ceil, floor


class RoundingScheme(Enum):
    floor = "floor"
    ceil = "ceil"
    near = "Near"


def round_to_int(
    value: float, scheme: RoundingScheme = RoundingScheme.floor
) -> int:
    if scheme == RoundingScheme.floor:
        return floor(value)
    elif scheme == RoundingScheme.ceil:
        return ceil(value)
    elif scheme == RoundingScheme.near:
        return round(value)
    else:
        print(
            "tried rounding with a invalid scheme (%s) using floor instead",
            scheme,
        )
        return int(floor(value))



from asyncio import run
from asyncio.protocols import Protocol

from sgr_specification.v0.generic import DataDirectionProduct

from sgr_commhandler.api.data_point_api import DataPoint
from sgr_commhandler.api.data_types import DataTypes


class FunctionalProfile(Protocol):
    def name(self) -> str: ...

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]: ...

    def get_data_point(self, dp_name: str) -> DataPoint:
        return self.get_data_points()[(self.name(), dp_name)]

    async def get_value_async(self) -> dict[str, DataPoint]:
        return {
            key[1]: await dp.get_value_async()
            for key, dp in self.get_data_points().items()
        }

    def get_value(self) -> dict[str, DataPoint]:
        return run(self.get_value_async())

    def describe(
        self,
    ) -> tuple[str, dict[str, tuple[DataDirectionProduct, DataTypes]]]:
        infos = map(lambda dp: dp.describe(), self.get_data_points().values())
        return self.name(), {dp[0][1]: (dp[1], dp[2]) for dp in infos}



from asyncio import run
from collections.abc import Callable
from typing import Any, Generic, Protocol, TypeVar

from sgr_specification.v0.generic import DataDirectionProduct

from sgr_commhandler.api.data_types import DataTypes

T = TypeVar('T')


class DataPointValidator(Protocol):
    def validate(self, value: Any) -> bool: ...

    def data_type(self) -> DataTypes: ...

    def options(self) -> list[Any] | None:
        return None


class DataPointProtocol(Protocol):
    async def set_val(self, value: Any): ...

    async def get_val(self, skip_cache: bool = False) -> Any: ...

    def name(self) -> tuple[str, str]: ...

    def direction(self) -> DataDirectionProduct: ...

    def can_subscribe(self) -> bool:
        return False

    def subscribe(self, fn: Callable[[Any], None]):
        raise Exception('Unsupported operatioin')

    def unsubscribe(self):
        raise Exception('Unsupported operatioin')


class DataPoint(Generic[T]):
    def __init__(
        self, protocol: DataPointProtocol, validator: DataPointValidator
    ):
        self._protocol = protocol
        self._validator = validator

    def name(self) -> tuple[str, str]:
        return self._protocol.name()

    async def get_value_async(self) -> T:
        value = await self._protocol.get_val()
        if self._validator.validate(value):
            return value
        raise Exception(
            f'invalid value read from device, {value}, validator: {self._validator.data_type()}'
        )

    def get_value(self) -> T:
        return run(self.get_value_async())

    async def set_value_async(self, value: T):
        if self._validator.validate(value):
            return await self._protocol.set_val(value)
        raise Exception('invalid data to write to device')

    def set_value(self, value: T):
        return run(self.set_value_async(value))

    def subscribe(self, fn: Callable[[Any], None]):
        self._protocol.subscribe(fn)

    def unsubscribe(self):
        self._protocol.unsubscribe()

    def direction(self) -> DataDirectionProduct:
        return self._protocol.direction()

    def data_type(self) -> DataTypes:
        return self._validator.data_type()

    def describe(
        self,
    ) -> tuple[tuple[str, str], DataDirectionProduct, DataTypes]:
        return self.name(), self.direction(), self.data_type()

    def options(self) -> list[Any]:
        options = self._validator.options()
        if options is None:
            return []
        return options



from typing import Any

from pymodbus.payload import BinaryPayloadBuilder, BinaryPayloadDecoder
from sgr_specification.v0.product.modbus_types import ModbusDataType


class PayloadDecoder(BinaryPayloadDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def decode(self, modbus_type: ModbusDataType, byte_count: int):
        """
        :param modbus_type: 'int8', 'int8_u', 'int16', 'int16_u', 'int32', 'int32_u', 'int64', 'int64_u', 'float32', 'boolean', 'float64', 'string'
        """
        # TODO enum, date_time
        if modbus_type.int8:
            return self.decode_8bit_int()
        elif modbus_type.int8_u:
            return self.decode_8bit_uint()
        elif modbus_type.int16:
            return self.decode_16bit_int()
        elif modbus_type.int16_u:
            return self.decode_16bit_uint()
        elif modbus_type.int32:
            return self.decode_32bit_int()
        elif modbus_type.int32_u:
            return self.decode_32bit_uint()
        elif modbus_type.int64:
            return self.decode_64bit_int()
        elif modbus_type.int64_u:
            return self.decode_64bit_uint()
        elif modbus_type.float32:
            return self.decode_32bit_float()
        elif modbus_type.float64:
            return self.decode_64bit_float()
        elif modbus_type.boolean:
            return bool(self.decode_8bit_uint())
        elif modbus_type.string:
            return self.decode_string(byte_count)
        else:
            raise ValueError("No supported modbus data type")


class PayloadBuilder(BinaryPayloadBuilder):
    def __init__(self, *args, **kwarg):
        super().__init__(*args, **kwarg)

    def sgr_encode(
        self, value: Any, modbus_type: ModbusDataType
    ) -> "PayloadBuilder":
        """
        :param modbus_type: 'int8', 'int8_u', 'int16', 'int16_u', 'int32', 'int32_u', 'int64', 'int64_u', 'float32', 'float64', 'boolean', 'string'
        """
        # TODO enum, date_time
        if modbus_type == ModbusDataType.int8:
            self.add_8bit_int(int(value))
        elif modbus_type == ModbusDataType.int8_u:
            self.add_8bit_uint(int(value))
        elif modbus_type == ModbusDataType.int16:
            self.add_16bit_int(int(value))
        elif modbus_type == ModbusDataType.int16_u:
            self.add_16bit_uint(int(value))
        elif modbus_type == ModbusDataType.int32:
            self.add_32bit_int(int(value))
        elif modbus_type == ModbusDataType.int32_u:
            self.add_32bit_uint(int(value))
        elif modbus_type == ModbusDataType.int64:
            self.add_64bit_int(int(value))
        elif modbus_type == ModbusDataType.int64_u:
            self.add_64bit_uint(int(value))
        elif modbus_type == ModbusDataType.float32:
            self.add_32bit_float(float(value))
        elif modbus_type == ModbusDataType.float64:
            self.add_64bit_float(float(value))
        elif modbus_type == ModbusDataType.boolean:
            self.add_8bit_uint(bool(value))
        elif modbus_type == ModbusDataType.string:
            self.add_string(str(value))
        else:
            print('Unsupported modbus type "%s"', modbus_type)
        return self



from typing import Optional

from sgr_specification.v0.product import (
    ConfigurationList,
    ConfigurationListElement,
)


def build_configuration_parameters(params: Optional[ConfigurationList]):
    if params is None:
        return []
    return [
        ConfigurationParameter(x) for x in params.configuration_list_element
    ]


class ConfigurationParameter:
    def __init__(self, parameter: ConfigurationListElement):
        translation = parameter.configuration_description
        self.label = translation[0].label
        self.name = parameter.name
        self.type = parameter.data_type
        self.description = translation[0].text_element
        self.default_value = parameter.default_value

    def __str__(self):
        return f'{self.label} - {self.name}'

    def __repr__(self):
        return f'{self.label} - {self.name}'



import logging
from typing import Any

from sgr_specification.v0.generic import (
    DataDirectionProduct,
)
from sgr_specification.v0.generic import (
    DataPointBase as ContactDataPointSpec,
)
from sgr_specification.v0.product import (
    ContactFunctionalProfile as ContactFunctionalProfileSpec,
)
from sgr_specification.v0.product import (
    DeviceFrame,
)

from sgr_commhandler.api import (
    DataPoint,
    DataPointProtocol,
    FunctionalProfile,
    SGrBaseInterface,
)
from sgr_commhandler.validators import build_validator

logger = logging.getLogger(__name__)


def build_contact_data_point(
    data_point: ContactDataPointSpec,
    functional_profile: ContactFunctionalProfileSpec,
    interface: 'SGrContactInterface',
) -> DataPoint:
    protocol = ContactDataPoint(data_point, functional_profile, interface)
    data_type = None
    if data_point.data_point and data_point.data_point.data_type:
        data_type = data_point.data_point.data_type
    validator = build_validator(data_type)
    return DataPoint(protocol, validator)


class ContactDataPoint(DataPointProtocol):
    def __init__(
        self,
        dp_spec: ContactDataPointSpec,
        fp_spec: ContactFunctionalProfileSpec,
        interface: 'SGrContactInterface',
    ):
        self._dp_spec = dp_spec
        self._fp_spec = fp_spec

        self._fp_name = ''
        if (
            fp_spec.functional_profile is not None
            and fp_spec.functional_profile.functional_profile_name is not None
        ):
            self._fp_name = fp_spec.functional_profile.functional_profile_name

        self._dp_name = ''
        if (
            self._dp_spec.data_point is not None
            and self._dp_spec.data_point.data_point_name is not None
        ):
            self._dp_name = self._dp_spec.data_point.data_point_name

        self._interface = interface

    def name(self) -> tuple[str, str]:
        return self._fp_name, self._dp_name

    async def get_val(self, skip_cache: bool = False):
        raise Exception('Not implemented')

    async def set_val(self, value: Any):
        raise Exception('Not implemented')

    def direction(self) -> DataDirectionProduct:
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.data_direction is None
        ):
            raise Exception('missing data direction')
        return self._dp_spec.data_point.data_direction


class ContactFunctionalProfile(FunctionalProfile):
    def __init__(
        self,
        fp_spec: ContactFunctionalProfileSpec,
        interface: 'SGrContactInterface',
    ):
        self._fp_spec = fp_spec
        self._interface = interface

        raw_dps = []
        if (
            self._fp_spec.data_point_list
            and self._fp_spec.data_point_list.data_point_list_element
        ):
            raw_dps = self._fp_spec.data_point_list.data_point_list_element

        dps = [
            build_contact_data_point(dp, self._fp_spec, self._interface)
            for dp in raw_dps
        ]

        self._data_points = {dp.name(): dp for dp in dps}

    def name(self) -> str:
        if (
            self._fp_spec.functional_profile
            and self._fp_spec.functional_profile.functional_profile_name
        ):
            return self._fp_spec.functional_profile.functional_profile_name
        return ''

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        return self._data_points


class SGrContactInterface(SGrBaseInterface):
    """
    SmartGridready External Interface Class for Contact Protocols
    Note: we do not implement a complete driver here, because it is very application-dependent!
    """

    def __init__(
        self, frame: DeviceFrame
    ):
        self._inititalize_device(frame)

        if (
            self.device_frame.interface_list
            and self.device_frame.interface_list
            and self.device_frame.interface_list.contact_interface
        ):
            self._raw_interface = self.device_frame.interface_list.contact_interface
        else:
            raise Exception('No contact interface')
        desc = self._raw_interface.contact_interface_description
        if desc is None:
            raise Exception('No contact interface description')

        # TODO configure interface
        self.number_of_contacts = desc.number_of_contacts
        self.contact_stabilization_time = desc.contact_stabilisation_time_ms

        raw_fps = []
        if (
            self._raw_interface.functional_profile_list
            and self._raw_interface.functional_profile_list.functional_profile_list_element
        ):
            raw_fps = self._raw_interface.functional_profile_list.functional_profile_list_element
        fps = [ContactFunctionalProfile(profile, self) for profile in raw_fps]
        self.functional_profiles = {fp.name(): fp for fp in fps}

    def is_connected(self):
        return False

    async def disconnect_async(self):
        # TODO implement after "driver"
        pass

    async def connect_async(self):
        # TODO implement after "driver"
        pass



import logging
import threading
from abc import ABC
from typing import Any, Optional

from pymodbus.client import AsyncModbusSerialClient, AsyncModbusTcpClient
from pymodbus.client.base import ModbusBaseClient
from pymodbus.constants import Endian
from sgr_specification.v0.product.modbus_types import BitOrder, ModbusDataType

from sgr_commhandler.driver.modbus.payload_decoder import (
    PayloadBuilder,
    PayloadDecoder,
)

logger = logging.getLogger(__name__)


class SGrModbusClient(ABC):
    def __init__(self, endianness: BitOrder):
        self._lock = threading.Lock()
        self._client: Optional[ModbusBaseClient] = None
        self._byte_order: Endian = (
            Endian.BIG
            if endianness is None or endianness == BitOrder.BIG_ENDIAN
            else Endian.LITTLE
        )
        self._word_order: Endian = (
            Endian.LITTLE
            if endianness
            in {BitOrder.CHANGE_WORD_ORDER, BitOrder.CHANGE_DWORD_ORDER}
            else Endian.BIG
        )

    async def connect(self): ...

    async def disconnect(self): ...

    def is_connected(self) -> bool: ...

    async def write_holding_registers(
        self, slave_id: int, address: int, data_type: ModbusDataType, value: Any
    ) -> None:
        """
        Encodes value to be written to holding register address
        :param slave_id: The slave ID of the device
        :param address: The address to read from and decode
        :param data_type: The modbus type to encode
        :param value: The value to be written
        """
        if self._client is None:
            raise Exception('Client not initialized')
        builder = PayloadBuilder(
            byteorder=self._byte_order, wordorder=self._word_order
        )
        builder.sgr_encode(value, data_type)
        with self._lock:
            await self._client.write_registers(
                address=address, values=builder.to_registers(), unit=slave_id
            )

    async def write_coils(
        self, slave_id: int, address: int, data_type: ModbusDataType, value: Any
    ) -> None:
        """
        Encodes value to be written to coil address
        :param slave_id: The slave ID of the device
        :param address: The address to read from and decode
        :param data_type: The modbus type to encode
        :param value: The value to be written
        """
        if self._client is None:
            raise Exception('Client not initialized')
        builder = PayloadBuilder(
            byteorder=self._byte_order, wordorder=self._word_order
        )
        builder.sgr_encode(value, data_type)
        with self._lock:
            await self._client.write_coils(
                address=address, values=builder.to_coils(), unit=slave_id
            )

    async def read_input_registers(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads input registers and decodes the value.
        :param slave_id: The slave ID of the device
        :param address: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded value
        """
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            response = await self._client.read_input_registers(
                address, count=size, slave=slave_id
            )
        if response and not response.isError():
            decoder = PayloadDecoder.fromRegisters(
                response.registers,
                byteorder=self._byte_order,
                wordorder=self._word_order,
            )
            return decoder.decode(data_type, 0)

    async def read_holding_registers(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads holding registers and decodes the value.
        :param slave_id: The slave ID of the device
        :param address: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded value
        """
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            response = await self._client.read_holding_registers(
                address, count=size, slave=slave_id
            )
        if response and not response.isError():
            decoder = PayloadDecoder.fromRegisters(
                response.registers,
                byteorder=self._byte_order,
                wordorder=self._word_order,
            )
            return decoder.decode(data_type, 0)

    async def read_coils(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads coils and decodes the value.
        :param slave_id: The slave ID of the device
        :param address: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded value
        """
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            response = await self._client.read_coils(
                address, count=size, slave=slave_id
            )
        if response and not response.isError():
            decoder = PayloadDecoder.fromCoils(
                response.bits,
                byteorder=self._byte_order,
                _wordorder=self._word_order,
            )
            return decoder.decode(data_type, 0)

    async def read_discrete_inputs(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads discrete inputs and decodes the value.
        :param slave_id: The slave ID of the device
        :param address: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded value
        """
        raise Exception('Discrete inputs not supported yet')

    # TODO Implement block transfers and remove this method
    async def _mult_value_decoder(
        self,
        addr: int,
        size: int,
        data_type: str,
        register_type: str,
        slave_id: int,
    ) -> Optional[float]:
        """
        Reads register and decodes the value.
        :param addr: The address to read from and decode
        :param size: The number of registers to read
        :param data_type: The modbus type to decode
        :returns: Decoded float
        """
        if self._client is None:
            raise Exception('Client not initialized')
        if register_type == 'HoldRegister':
            reg = self._client.read_holding_registers(
                addr, size, slave=slave_id
            )
        else:
            reg = self._client.read_input_registers(
                addr, count=size, slave=slave_id
            )
        reg = await reg
        decoder = PayloadDecoder.fromRegisters(
            reg.registers,
            byteorder=self._byte_order,
            wordorder=self._word_order,
        )
        # logger.debug(decoder.decode('float32', 0))
        if not reg.isError():
            # logger.debug(decoder.decode(data_type, 0))
            indexes = [size // 3 * 0, size // 3 * 1, size // 3 * 2]
            l1 = decoder.decode(data_type, indexes[0])
            l2 = decoder.decode(data_type, indexes[1])
            l3 = decoder.decode(data_type, indexes[2])
            return l1, l2, l3


class SGrModbusTCPClient(SGrModbusClient):
    def __init__(self, ip: str, port: int, endianness: BitOrder):
        super().__init__(endianness)
        """
        Creates client
        :param ip: The host to connect to (default 127.0.0.1)
        :param port: The modbus port to connect to (default 502)
        """
        self._ip = ip
        self._port = port
        self._client = AsyncModbusTcpClient(
            host=ip,
            port=port,
            timeout=1,
            retries=0,
            reconnect_delay=5000,
            reconnect_delay_max=30000,
        )

    async def connect(self):
        if self._client is None:
            raise Exception('Client not initialized')

        with self._lock:
            await self._client.connect()
            logger.debug('Connected to ModbusTCP on ip: ' + self._ip)

    async def disconnect(self):
        if self._client is None:
            return
        with self._lock:
            self._client.close()
            logger.debug('Disconnected from ModbusTCP on ip: ' + self._ip)

    def is_connected(self) -> bool:
        return self._client is not None and self._client.connected


class SGrModbusRTUClient(SGrModbusClient):
    def __init__(
        self, serial_port: str, parity: str, baudrate: int, endianness: BitOrder
    ):
        super().__init__(endianness)
        """
        Creates client
        :param serial_port: The serial port to connect to (e.g. COM1)
        :param parity: The serial parity (e.g. EVEN)
        :param baudrate: The serial baudrate (e.g. 19200)
        """
        self._serial_port = serial_port
        self._client = AsyncModbusSerialClient(
            method='rtu',
            port=serial_port,
            parity=parity,
            baudrate=baudrate,
        )  # changed source: https://stackoverflow.com/questions/58773476/why-do-i-get-pymodbus-modbusioexception-on-20-of-attempts

    async def connect(self):
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            _is_connected = await self._client.connect()
            logger.debug(
                'Connected to ModbusRTU on serial port: ' + self._serial_port
            )

    async def disconnect(self):
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            self._client.close(reconnect=False)
            logger.debug(
                'Disconnected from ModbusRTU on serial port: '
                + self._serial_port
            )

    def is_connected(self) -> bool:
        return self._client is not None and self._client.connected
